;-*- Mode: Common-lisp; Package: optop; Readtable: ytools; -*-
(in-package :optop)
;;;$Id: execute.lisp,v 1.29 2006/01/12 19:24:16 dvm Exp $

(depends-on %module/ nisp)

;;; Functions for executing and projecting action expansions
;;; This file depends hardly at all on the rest of Optop; it is a
;;; general-purpose nondeterministic interpreter, and it knows nothing
;;; of the syntax of Opt.  (The only concession is to make Exec-bdg-id's
;;; explicitly include Plan-links.)

(depends-on :at-run-time %optop/ types projutil %ytools/ nilscompat)

(depends-on :at-compile-time %optop/ copy-w-changes next-exec-state)

;;; The code from forceproj.lisp has been turned inside out (oops, I mean 
;;; "refactored") and subsumed under execution in :forced mode.

;;; Questions: Which comes first when you're executing both an
;;; expansion and a reduction tree?  (Answer: The reduction tree,
;;; because it can achieve goals the expansion set up.)

(defnisptype Exec-state :forward)

;;; :find-open-subgoals is the mode used in 'make-or-edit-rm-graph' to find
;;; gaps that an rm-graph can fill.

(defnisptype Exec-mode (Const :real :projected :forced :one-step :zero-step))
;;; :one-step and :zero-step are not forced.  See below.

(defnisptype Exec-status (Const :finished :continuing :ready
				:blocked :stalled :failed))
;;; :failed -- Infeasible step attempted
;;; :continuing -- Okay so far, go on to next step (= Projection-result :okay)
;;; :ready -- In :zero-step mode, means thread reached point where the next
;;;               thing is to take an action.
;;; :blocked -- All threads blocked waiting for something to change 
;;; :stalled -- Exceeded resource allotment
;;; :finished -- Done, quiescent, hooray

;;; The status flags above can be returned by individual thread continuations
;;; as well as by 'continue-execution' itself.  But most of them make no
;;; sense when returned by a continuation.  All are treated as "comments,"
;;; equivalent to :continuing, except for :failed.  E.g., :finished means
;;; "this thread is done"; :blocked means "I put myself on the blockage
;;; queue"; and so forth.

(defnisptype Thread-stack :forward)

;;; These are just for debugging; the Obj's are just going to be printed.
(defnisptype Nonempty-thread-stack (Structure this-frame - Obj
					      rest - Thread-stack))

(eval-when (:compile-toplevel :load-toplevel :slurp-toplevel :execute)
   (defvar empty-thread-stack* (printable-as-string "<Empty thread stack>")))

(defnisptype Thread-stack
             (Alt Nonempty-thread-stack (Const #.empty-thread-stack*))
   (:access is-empty - Boolean :inline (ts - Thread-stack)
       (eq ts empty-thread-stack*))
   (:conser (frame - Obj s - Thread-stack)
       (make Nonempty-thread-stack frame s))
)

(specdecl empty-thread-stack* - Thread-stack)

;;; Result returned by every Thread-cont
(defnisptype TC-val (Multvals status - Exec-status
			      state - Exec-state))

;;; These actually do some work, and report the results as a TC-val.
(defnisptype Thread-cont (Fun TC-val
			      (Exec-state Thread-stack)
			      t))

(specdecl (no-op-thread-cont* (\\ (e - Exec-state _ - Thread-stack)
				 (values (be Exec-status ':continuing)
                                         e)))
	  - Thread-cont)


;;; (tg $c$) is a continutation that does something, then calls $c$.
(defnisptype Thread-gen (Fun Thread-cont (Thread-cont)))

(specdecl (no-op-thread-gen*
	     (be Thread-gen (\\ - Thread-cont (tc - Thread-cont) tc)))
	  - Thread-gen)

(defnisptype Thread (Structure continuation - Thread-cont
			       stack - Thread-stack))

(deffunc new-Thread (cont frame stack)
   (make Thread cont (make Thread-stack frame stack)))

;;; The type of !_data depends on !_kind, which is either :open-subgoals,
;;; :link, or :ready, or ...?
;;; ':ready' means not really blocked, but don't want to proceed.
;;;          -> data = nil
;;; ':open-subgoals' -> data = (subgoal in dnf (Lst (Lst Sexp)), continuation)
;;;        continuation is function that takes list of answers to subgoal,
;;;        a list of Threads (= (!_threads blockage)),
;;;        and returns a list of Threads.  
(defnisptype Gate (nisp:Object kind - Symbol data - Obj))

(defnisptype Blockage (Structure gate - Gate threads - (Lst Thread)))

;;; How bindings work: Because we can't destructively alter Exec-states
;;; (defined below), we can't assign a value cell to a newly bound
;;; variable.  Instead, we assign it an Exec-bdg-id, and then store
;;; the values of the id in the !_vals slot of Exec-states in which
;;; the variable binding is in effect.  

(defnisptype Exec-bdg-id (Alt Symboid Plan-link)
   (conser :inline (str) (make-Symboid str)))

;;; Records list of values produced by the last step -- 
(specdecl value-id* (build-Symboid :exec-val*))

(defnisptype Exec-val-tab (Lst (Lrcd Exec-bdg-id Obj)))
(defnisptype Exec-bdg-tab (Lst (Lrcd CSymbol Exec-bdg-id)))

;;; The fields of this datatype must never be set, because the
;;; same Exec-state can occur in different branches of the search space.
;;; Instead use 'next-exec-state' to copy current state with some changes.
(defnisptype Exec-state (Structure mode - Exec-mode
				   ;;;;forcings
				   steps - Integer
				   elapsed - (Alt Time-int (Const *never))
				   ;; -- *never only if mode is :failed?
				   active - (Lst Thread)
				   pp - Plan-prefix
				   sit-space - Situation-space
				   hist - (Lst Exec-state)
				    ;; -- In non-:real modes, the situations
				    ;; and execution states
				    ;;; leading up to this one,
				    ;; in rev chrono order.
				   bdgs - Exec-bdg-tab
				   vals - Exec-val-tab
				   blocked - (Lst Blockage)
;;; Surely this is subsumed by '!_blocked' --
;;;;				   ;; Goals derived from (achieve p)
;;;;				   ;; steps --
;;;;				   unachieved - (Lst (Lrcd Occasion
;;;;							   Situation))
				   )
;;;;  (start-sit - Situation (es - Exec-state)
;;;;     (car (car (!_hist es))))
  (past-sits - (Lst Situation) (es - Exec-state)
     (<# !_sit (!_hist es)))

  (domain - Domain (es - Exec-state)
     (!_domain (!_initsit (!_sit-space es))))

  (sit - Situation (es - Exec-state)
     (!_sit (!_pp es)))

  (env - Exec-bdg-tab (es - Exec-state) (!_bdgs es))
)

;;;;Have to fix opt-eval so it expects args giving Exec-bdg/val-tab.

;;;;How do you evaluate during real execution?
;;;;E.g., need some notion of "the current actual situation."

(specdecl (interleave-threads* false) - Boolean)

(specdecl (max-time-link* (new-Plan-link ':max-time-left))
	  (max-steps-link* (new-Plan-link ':max-steps-left))
	  - Plan-link)

(defproc continue-execution - (Multvals status - Exec-status
					new-state - Exec-state
					rem-cycles - Integer)
			    (state - Exec-state
			     max-time - Time-int max-steps - Integer 
			     cycles - Integer)
   (!= state
       (next-exec-state *-*
	  (:vals
	     (exec-value-set
	         max-time-link*
		 (exec-value-set max-steps-link* !>vals
				 false (\\ () max-steps))
		 false
		 (\\ () max-time)))))
   (repeat :for ((status ':continuing) - Exec-status
		 (other-threads !()) active-threads - (Lst Thread))
      (!= state (check-subgoal-blockages state))
      (!= active-threads (!_active state))
    :until (or (not (eq status ':continuing))
	       (null active-threads)
	       (= cycles 0)
	       (> (!_steps state) max-steps)
	       (> (be Time-int (!_elapsed state)) max-time))
    :result (values (cond ((null active-threads)
			   (cond ((null (!_blocked state))
				  ':finished)
				 (t ':blocked)))
			  ((eq status ':continuing)
			   ':stalled)
			  ;; -- This is the *only* way to return a :stalled
			  ;; status.  Threads never return it.
			  (t status))
		    state cycles)
      ;; 'other-threads' is the list of threads that were not picked
      ;; on the previous iteration (empty if there was just one thread).
      ;; This is to ensure that parallel threads get equal time,
      ;; but it probably screws up plans by planner. --
      (cond ((and interleave-threads*
		  (not (null other-threads))
		  (tailp other-threads active-threads))
	     (!= active-threads
		 (append other-threads
			 (ldiff active-threads other-threads)))))
      (!= other-threads (tail active-threads))
      (!= < status state >
	  (funcall (!_continuation (head active-threads))
		   (next-exec-state state
		      (:active other-threads
		       :hist (cons state !>hist)))
		   (!_stack (head active-threads))))
      (cond ((not (memq status '(:continuing :failed)))
	     (!= status ':continuing)))
      (!= cycles (- *-* 1))))

;;; This version does a nondeterministic simulation just one step (situation
;;; change) ahead.  It returns all the possible actions that could be executed
;;; next.
(defproc execute-one-step - (Lst (Lrcd (~ Action) Exec-status Exec-state))
	                  (state - Exec-state)
   (!= state (check-subgoal-blockages state))
   (repeat :for (;;;;(initsit (!_sit state)) - Situation
		 ;; Managed as a stack --
		 (expansion-queue !()) - (Lst Exec-state)
		 :collector results - (Lrcd (~ Action) Exec-status Exec-state))
    :nconc
      (repeat :for ((th :in (!_active state)) - Thread
		    new-status - Exec-status new-state - Exec-state)
	 (!= < new-status new-state >
	     (funcall (!_continuation th)
		      (next-exec-state state
			 (:active (remove1q th (!_active state))
			  :hist (cons state !>hist)))
		      (!_stack th)))
       :within
         (let ((act (cond ((eq (!_sit new-state)
			       (!_sit state))
			   false)
			  (t (between-action (!_sit state)
					     (!_sit new-state)
					     (!_sit-space state)))))
	       - (~ (Alt Action (Const wait transition))))
	    (cond ((and act (not (memq act '(wait transition))))
		   (let ((e (assq act (!_elements results))))
		      (cond (e
			     (cond ((exec-status-better new-status (cadr e))
				    (!= (cadr e) new-status)
				    (!= (caddr e) new-state))))
			    (t
			     (:continue
			      :collect (tuple act new-status new-state))))))
		  ((eq new-status ':finished)
		   (:continue
		    :collect (tuple false ':finished new-state)))
		  ((eq new-state ':continuing)
		   (on-list new-state expansion-queue)))))
      :until (null expansion-queue)
         (!= state (off-list expansion-queue))
      :result results))

;;; Execute as far as possible without actually doing anything.  This will
;;; expose all the open subgoals and other blockages.
(deffunc exec-state-normalize - Exec-state (es0 - Exec-state)
   (let ()
      (repeat :for ((es (check-subgoal-blockages es0)) - Exec-state
		    ;;;; status - Exec-status
		    )
       :until (null (!_active es))
	 (!= < _ es >
	     (funcall (!_continuation (car (!_active es)))
		      (copy-with-changes es
			 (:active (cdr !>active)
			  :hist (cons es !>hist)
			  :mode ':zero-step))
		      (!_stack (car (!_active es)))))
       :result (multi-let (((ready others - (Lst Blockage))

			    (classify (!_blocked es)
				      (\\ (b - Blockage)
					 (eq (!_kind (!_gate b))
					     ':ready)))))
		  (copy-with-changes es 
		     (:active (append (<$ !_threads ready)
				      (!_active es))
		      :blocked others))))))

(deffunc exec-status-better - Boolean (stat1 stat2 - Exec-status)
   (memq stat1
	 (cdr (memq stat2 '(:failed :blocked :continuing :finished)))))

(deffunc between-action - Action
	      (sit1 sit2 - Situation sit-space - Situation-space)
   (repeat :for ((ss :in (!_succs sit1 sit-space)))
    :result (progn
	       (dbg-save :run-loud sit1 sit2)
	       (signal-problem between-action
		  "Can't find action leading from " sit1 " to " sit2))
    :until (and (eq (!_sit ss) sit2)
		(is Action (!_act ss)))
    :result (be Action (!_act ss))))

(specdecl (check-id* (new-Varid)) - Varid)

;;; Free up threads for subgoal blockages that are now unblocked
;;; (because the subgoals are now true).
(deffunc check-subgoal-blockages - Exec-state (state - Exec-state)
   (let ((sit (!_sit state)))
      (repeat :for ((b :in (!_blocked state)) - Blockage
		    (dom (!_domain (find-init-situation sit))) - Domain
		    :collector keep - Blockage free - Thread)
       :within
	 (cond ((eq (!_kind (!_gate b)) ':open-subgoals)
		(decl ((d (!_data (!_gate b)))
		       - (Lrcd Sexp (Fun (Lst Thread)
					 ((Lst Bdgenv) (Lst Thread)))))
		   (let ((al (deduce `(or ,@(<# (\\ (c) `(and ,@c))
						(car d)))
				     check-id* (empty-env) sit dom)))
		      (cond ((null al)
			     (:continue
			      :collect (:into keep b)))
			    (t
			     ;; Notify the thread that 
			     (:continue
			      :nconc (:into free
					    (funcall (cadr d)
						     al (!_threads b)))))))))
	       (t
		(:continue
		 :collect (:into keep b))))
       :result (cond ((null free)
		      state)
		     (t
		      (copy-with-changes state
			 (:active (nconc free !>active)
			  :blocked keep)))))))


(specdecl (plan-id* (new-Varid)) - Varid)

(defnisptype Projection-result (Alt (Const :failed)
				    ;; Looking for open subgoals,
				    ;; but they're not here --
				    (Const :ready)
				    (Lstructure &flag :okay
				       sit - Situation
				       elapsed - Time-int
				       numsteps - Integer)
				    (Lstructure &flag :forced
				       num-forcings - Integer
				       sit - Situation
				       elapsed - Time-int
				       numsteps - Integer)
				    (Lstructure &flag :open-subgoals
				       dnf - (Lst (Lst Sexp)))))

;;; If result is false, that means there is no useful cached transition.
(deffunc recover-cached-transition - (~ Projection-result)
				   (mode - Exec-mode
				    sit - Situation act - Action
				    sit-space - Situation-space)
   (let ((sss (find-old-projection sit act sit-space))
	 - (~ Sitsuccspec))
      (cond (sss
	     (with sss
		(cond ((null !>open-subgoals)
		       (cond (!>sit
			      (cond ((= !>forcings 0)
				     (tuple (atomconst :okay)
					    !>sit
					    (be Time-int !>elapsed)
					    !>steps))
				    ((eq mode ':forced)
				     (tuple (atomconst :forced)
					    !>forcings
					     !>sit
					     (be Time-int !>elapsed)
					     !>steps))
				    (t ':failed)))
			     (t ':failed)))
		      (t
		       (cond (!>sit
			      (cond ((= !>forcings 0)
				     (signal-problem recover-cached-transition
					"Unforced execution of " act
					" in spite of open subgoals "
					:% " [to wit -- "
					!>open-subgoals "]"))
				    ((eq mode ':projected)
				     ':failed)
				    (t
				     (tuple (atomconst :forced)
					    !>forcings
					    !>sit
					    (be Time-int !>elapsed)
					    !>steps))))
			     (t false))))))
	    (t false))))

(deffunc exec-value-get - Obj (id - Exec-bdg-id alist - Exec-val-tab)
   (alref alist id (signal-problem exec-value-get
		      "Accessing missing value of " id)))

;;; This does _not_ alter alist, but produces a new alist with
;;; the old binding replaced with a new one whose 
;;; cadr = (val-computer old-val).
;;; If there is no old binding, an error is signaled unless 'val-init'
;;; is non-false, in which case it is used to produce a new binding.
;;; If 'val-computer' is false, the old value is ignored and 'val-init'
;;; is always used to set the value.  
;;; The location of the resulting binding in the alist is unpredictable
;;; and subject to change.
(deffunc exec-value-set - Exec-val-tab
	       (id - Exec-bdg-id alist - Exec-val-tab
		val-computer - (~ (Fun Obj (Obj)))
		val-init - (~ (Fun Obj ())))
   (let ((e (assq id alist)))
      (cond (e
	     (let ((new-val 
		      (cond (val-computer
			     (funcall val-computer (second e)))
			    (val-init
			     (funcall val-init))
			    (t
			     (signal-problem link-value-set
				"Storing into " id " with no "
				"way to compute value")))))
		(cons (tuple id new-val)
		      (remove1q e alist))))
	    (val-init
	     (let ((new-val (funcall val-init)))
		(cons (tuple id new-val)
		      alist)))
	    (t
	     (signal-problem link-value-set
		"Initializing " id " with no way to compute value")))))

(deffunc exec-value-unbind - Exec-val-tab
	          (id - Exec-bdg-id alist - Exec-val-tab)
   (let ((e (alref alist id)))
      (cond (e
	     (remove1q e alist))
	    (t alist))))

(deffunc links-unblock - Exec-state (links - (Lst Plan-link)
				     estate - Exec-state)
   (repeat :for ((b :in (!_blocked estate)) - Blockage
		 :collectors remaining-blocks - Blockage
		             new-threads - Thread)
    :within
       (cond ((and (eq (!_kind (!_gate b)) ':link)
		   (memq (!_data (!_gate b)) links))
	      (:continue
	       :append (:into new-threads (!_threads b))))
	     (t
	      (:continue
	       :collect (:into remaining-blocks b))))
     :result (copy-with-changes estate
		(:blocked remaining-blocks
		 :active (nconc new-threads !>active)))))

(deffunc cont-trampoline - Thread-cont (tc - Thread-cont)
   (\\ (estate - Exec-state tstack - Thread-stack)
       - TC-val
      (thread-enqueue
         (thread-cont-encapsulate tc tstack)
	 estate)))
;;; -- has the same execution semantics as 'tc' itself, 
;;; but "trampolines" back to execution engine ('continue-execution').

(deffunc thread-cont-encapsulate - Thread (c - Thread-cont stk - Thread-stack)
   (make Thread
	 (\\ (est - Exec-state _ - Thread-stack)
	     - (Mlv Exec-status Exec-state)
	    (funcall c est stk))
	 stk))

(deffunc thread-enqueue - TC-val
	     (thr - Thread estate - Exec-state)
   (continuing-exec
      (copy-with-changes estate
	 (:active (cons thr !>active)))))

(deffunc continuing-exec - TC-val (estate - Exec-state)
   (values ':continuing estate))

;;; THIS SHOULD BE CALLED BY force-project

(defproc forced-action-debug - Void (act - Action
				     goal - (~ Sexp) goal-achieved - Boolean
				     forced - Boolean
				     steps - Integer
				     elapsed - (Alt Time-int (Const *never)))
   (out (:to *error-output*)
	    (:q (forced  
		 "Action " act :% " infeasible" :% " FORCING" :%))
	    (:q (goal
		 (:q (goal-achieved
		      "Okay, goal now satisfied")
		     (t
		      "Goal not satisfied"))))
	    1
	    (:q ((not (eql elapsed 0.0))
		 "Elapsed = " elapsed 1))
	    (:q ((not (eql steps 0))
		 "Steps = " steps 1))
	    :%))